/* 
 * trans.c - Matrix transpose B = A^T
 *
 * Each transpose function must have a prototype of the form:
 * void trans(int M, int N, int A[N][M], int B[M][N]);
 *
 * A transpose function is evaluated by counting the number of misses
 * on a 1KB direct mapped cache with a block size of 32 bytes.
 */ 
#include <stdio.h>
#include "cachelab.h"

int is_transpose(int M, int N, int A[N][M], int B[M][N]);

/* 
 * transpose_submit - This is the solution transpose function that you
 *     will be graded on for Part B of the assignment. Do not change
 *     the description string "Transpose submission", as the driver
 *     searches for that string to identify the transpose function to
 *     be graded. 
 */
void transpose_32x32(int M, int N, int A[N][M], int B[M][N]) {
    // fprintf(stderr, "%p %p\n", A, B);
    for (int ii = 0; ii < 32; ii += 8)
        for (int jj = 0; jj < 32; jj += 8)
            for (int i = ii; i < ii + 8; ++i) {
                int a0 = A[i][jj];
                int a1 = A[i][jj + 1];
                int a2 = A[i][jj + 2];
                int a3 = A[i][jj + 3];
                int a4 = A[i][jj + 4];
                int a5 = A[i][jj + 5];
                int a6 = A[i][jj + 6];
                int a7 = A[i][jj + 7];
                B[jj][i] = a0;
                B[jj + 1][i] = a1;
                B[jj + 2][i] = a2;
                B[jj + 3][i] = a3;
                B[jj + 4][i] = a4;
                B[jj + 5][i] = a5;
                B[jj + 6][i] = a6;
                B[jj + 7][i] = a7;
            }
}
void transpose_64x64(int M, int N, int A[N][M], int B[M][N]) {
    int a0, a1, a2, a3, a4, a5, a6, a7;
    for (int ii = 0; ii < N; ii += 8)
        for (int jj = 0; jj < M; jj += 8) {
            for (int i = ii; i < ii + 4; ++i) {
                a0 = A[i][jj + 0];
                a1 = A[i][jj + 1];
                a2 = A[i][jj + 2];
                a3 = A[i][jj + 3];
                a4 = A[i][jj + 4];
                a5 = A[i][jj + 5];
                a6 = A[i][jj + 6];
                a7 = A[i][jj + 7];
                B[jj + 0][i] = a0;
                B[jj + 1][i] = a1;
                B[jj + 2][i] = a2;
                B[jj + 3][i] = a3;
                B[jj + 0][i + 4] = a4;
                B[jj + 1][i + 4] = a5;
                B[jj + 2][i + 4] = a6;
                B[jj + 3][i + 4] = a7;
            }
            for (int j = jj; j < jj + 4; ++j) {
                a0 = B[j][ii + 4];
                a1 = B[j][ii + 5];
                a2 = B[j][ii + 6];
                a3 = B[j][ii + 7];
                B[j][ii + 4] = A[ii + 4][j];
                B[j][ii + 5] = A[ii + 5][j];
                B[j][ii + 6] = A[ii + 6][j];
                B[j][ii + 7] = A[ii + 7][j];
                B[j + 4][ii + 0] = a0;
                B[j + 4][ii + 1] = a1;
                B[j + 4][ii + 2] = a2;
                B[j + 4][ii + 3] = a3;
            }
            for (int i = ii + 4; i < ii + 8; ++i) {
                a0 = A[i][jj + 4];
                a1 = A[i][jj + 5];
                a2 = A[i][jj + 6];
                a3 = A[i][jj + 7];
                B[jj + 4][i] = a0;
                B[jj + 5][i] = a1;
                B[jj + 6][i] = a2;
                B[jj + 7][i] = a3;
            }
        }
}
void transpose_61x67(int M, int N, int A[N][M], int B[M][N]) {
    int a0, a1, a2, a3, a4, a5, a6, a7;
    for (int ii = 0; ii < N; ii += 8)
        for (int jj = 0; jj < M; jj += 8)
            for (int i = ii; i < ii + 8 && i < N; ++i) {
                if (jj + 0 < M) a0 = A[i][jj];
                if (jj + 1 < M) a1 = A[i][jj + 1];
                if (jj + 2 < M) a2 = A[i][jj + 2];
                if (jj + 3 < M) a3 = A[i][jj + 3];
                if (jj + 4 < M) a4 = A[i][jj + 4];
                if (jj + 5 < M) a5 = A[i][jj + 5];
                if (jj + 6 < M) a6 = A[i][jj + 6];
                if (jj + 7 < M) a7 = A[i][jj + 7];
                if (jj + 0 < M) B[jj][i] = a0;
                if (jj + 1 < M) B[jj + 1][i] = a1;
                if (jj + 2 < M) B[jj + 2][i] = a2;
                if (jj + 3 < M) B[jj + 3][i] = a3;
                if (jj + 4 < M) B[jj + 4][i] = a4;
                if (jj + 5 < M) B[jj + 5][i] = a5;
                if (jj + 6 < M) B[jj + 6][i] = a6;
                if (jj + 7 < M) B[jj + 7][i] = a7;
            }
}
char transpose_submit_desc[] = "Transpose submission";
void transpose_submit(int M, int N, int A[N][M], int B[M][N])
{
    if (M == 32 && N == 32) transpose_32x32(M, N, A, B);
    else if (M == 64 && N == 64) transpose_64x64(M, N, A, B);
    else transpose_61x67(M, N, A, B);
}

/* 
 * You can define additional transpose functions below. We've defined
 * a simple one below to help you get started. 
 */ 

/* 
 * trans - A simple baseline transpose function, not optimized for the cache.
 */
char trans_desc[] = "Simple row-wise scan transpose";
void trans(int M, int N, int A[N][M], int B[M][N])
{
    int i, j, tmp;

    for (i = 0; i < N; i++) {
        for (j = 0; j < M; j++) {
            tmp = A[i][j];
            B[j][i] = tmp;
        }
    }    

}

/*
 * registerFunctions - This function registers your transpose
 *     functions with the driver.  At runtime, the driver will
 *     evaluate each of the registered functions and summarize their
 *     performance. This is a handy way to experiment with different
 *     transpose strategies.
 */
void registerFunctions()
{
    /* Register your solution function */
    registerTransFunction(transpose_submit, transpose_submit_desc); 

    /* Register any additional transpose functions */
    registerTransFunction(trans, trans_desc); 

}

/* 
 * is_transpose - This helper function checks if B is the transpose of
 *     A. You can check the correctness of your transpose by calling
 *     it before returning from the transpose function.
 */
int is_transpose(int M, int N, int A[N][M], int B[M][N])
{
    int i, j;

    for (i = 0; i < N; i++) {
        for (j = 0; j < M; ++j) {
            if (A[i][j] != B[j][i]) {
                return 0;
            }
        }
    }
    return 1;
}

